"""pil_frame_reader - module for PIL-based functionality"""
from pathlib import Path
import os
import numpy as np
from vre_video.utils import natsorted, image_read
from .frame_reader import FrameReader

class PILFrameReader(FrameReader):
    """implements FrameReader using Pillow to read frame by frame"""
    def __init__(self, path: str | Path, fps: float | None = None, frames: list[int] | None = None):
        self._path = Path(path)
        self.frame_paths = natsorted(list(self._path.iterdir()), key=lambda p: p.name)
        self.data: list[np.ndarray] = [image_read(x) for x in self.frame_paths]
        _shp = self.data[0].shape
        assert all(x.shape == _shp for x in self.data), f"Not all shapes of all images are equal to first image: {_shp}"
        assert all(x.dtype == np.uint8 for x in self.data), f"Not all data is uint8: {set(x.dtype for x in self.data)}"
        self.data = np.array(self.data)
        self._fps = fps or float(os.getenv("VIDEO_FPS", "1"))

        # try to make the frames based on 1.png, 2.png, etc.
        if frames is None:
            try:
                frames = sorted([int(x.stem) for x in self.frame_paths])
            except Exception:
                pass

        self.frames = frames or list(range(len(self.data)))
        self.frames_ix = dict(zip(self.frames, range(len(self.frames)))) # {ix: frame}

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return (len(self.data), *self.data[0].shape)

    @property
    def fps(self):
        return self._fps

    @property
    def path(self) -> str:
        return str(self._path)

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        if isinstance(ix, int):
            return self.data[self.frames_ix[ix]]
        if isinstance(ix, np.ndarray):
            ix = ix.tolist()
        if isinstance(ix, slice):
            ix = list(range(ix.start, ix.stop))
        ix_transformed = [self.frames_ix[_ix] for _ix in ix] # support for sparse frames (VRE)
        return self.data[ix_transformed]

    def __len__(self):
        return self.frames[-1] + 1



"""
fd_frame_reader - module for reading from a file descriptor (like stdin). To be used with external generating data
tool like ffmpeg (i.e. for webcams) and linux pipes.
"""
from io import IOBase
from threading import Lock, Thread
import numpy as np
from .frame_reader import FrameReader

from ..utils import logger

class FdFrameReader(FrameReader):
    """
    Implements FrameReader that reads from an io stream (like stdin). Can be used with linux pipes.
    Parameters:
    - data The socket/IO base we are reading from
    - resolution The resolution the raw data is read as
    - fps not really used
    - async_worker If set to false, then all the frames will be read. This is possibly a bad idea as we may process
    slower than they arive, leading to memory growth. Needed when we care about reading everything.
    """
    def __init__(self, data: IOBase, resolution: tuple[int, int], fps: int, async_worker: bool=True):
        assert isinstance(data, IOBase), type(data)
        super().__init__()
        self.data = data
        self._fps = fps
        self.resolution = (self.height, self.width) = resolution
        self.async_worker = async_worker

        self.lock = Lock()
        self.current_frame: np.ndarray | None = np.zeros((*resolution, 3), dtype=np.uint8)
        self.current_frame_ix = 0
        if self.async_worker:
            Thread(target=self._read_frame_worker, daemon=True).start()

    @property
    def shape(self) -> tuple[int, int, int, int]:
        return (2**31, self.height, self.width, 3)

    @property
    def fps(self):
        return self._fps

    @property
    def path(self) -> str:
        return str(self.data)

    def _read_one_frame(self) -> np.ndarray:
        raw_frame, frame = None, None
        try:
            raw_frame = self.data.read(self.width * self.height * 3)
        except ValueError as e:
            if str(e) not in ("I/O operation on closed file", "read of closed file"):
                raise e
        if raw_frame is not None and len(raw_frame) != 0:
            frame = np.frombuffer(raw_frame, dtype=np.uint8).reshape((self.height, self.width, 3))
        if self.current_frame is None or not np.allclose(frame, self.current_frame):
            self.current_frame_ix += 1
        return frame

    def _read_frame_worker(self):
        """note: we need to use a thread here otherwise ffmpeg is too fast for our sync reader"""
        while True:
            frame = self._read_one_frame()
            if frame is None:
                break
            with self.lock:
                self.current_frame = frame
        logger.debug("read frame worked closed")

    def __getitem__(self, ix: int | list[int] | np.ndarray | slice) -> np.ndarray:
        if self.async_worker is False:
            frame = self._read_one_frame()
        else:
            with self.lock:
                frame = self.current_frame

        if frame is None:
            raise StopIteration
        if isinstance(ix, slice):
            assert (ix.start - ix.stop) == 1, f"cannot have batches in fd reader, got: {list(ix)}"
            ix = [0]
        if isinstance(ix, (list, np.ndarray)):
            assert len(ix) == 1, f"cannot have batches in fd reader, got: {ix}"
            frame = frame[None] # this function always returns a batch of one for other tools (i.e. streaming)
        return frame



#!/usr/bin/env python3
"""
vre_video_player - a simple video player using tkinter
Usage: ./vre_video_player.py VIDEO [--hud_font_color 255 0 0] [--hud_font_size R G B] [--stats_path STATS.CSV]
                                    [--input_size H W]
If VIDEO is stdin ("-") or a socket (tcp://), '--input_size' must be set as it's passed to VREVideo.
"""
import tkinter as tk
from argparse import ArgumentParser, Namespace
from queue import Queue, Empty
from datetime import datetime
import time
import sys
import threading
from PIL import Image, ImageTk
import numpy as np

from vre_video.utils import logger, image_add_text
from vre_video import VREVideo

TIMEOUT_READ_S = 2
TIMEOUT_WRITE_S = 0.5

class VideoContainer:
    """Video Container class. Holds the state of the video player shared by the UI and the reading thread"""
    def __init__(self, video: VREVideo, font_size: float | None = None, font_color: tuple[int, int, int] | None = None):
        self.video = video
        self.fps_multiplier: float = 1.0
        self._current_frame: int = 0
        self.is_paused: bool = False
        self.font_color = tuple(font_color or (255, 0, 0))
        self.font_size = font_size or max(10, self.frame_shape[1] // 50)

    @property
    def fps(self) -> float:
        """The fps of the video"""
        return self.video.fps

    @property
    def frame_shape(self) -> tuple[int, int, int]:
        """The frame shape of the video"""
        return self.video.frame_shape

    @property
    def current_frame(self):
        """the current frame of the video"""
        return self._current_frame

    @current_frame.setter
    def current_frame(self, current_frame: int):
        if not 0 <= current_frame < len(self.video):
            logger.info(f"Frame {current_frame} wanted which is wrong ({len(self.video)=}). Resetting to 0.")
            current_frame = 0
        self._current_frame = int(current_frame)

class VideoPlayer(threading.Thread):
    """with the help of the chat-ge-pe-tees"""
    def __init__(self, queue: Queue, video_container: VideoContainer):
        super().__init__()
        self.queue = queue
        self.video_container = video_container
        self.show_stats = True
        self.frame_counter = 0
        self.fps_counter = np.full(100, dtype=np.float32, fill_value=float("nan"))
        self.root: tk.Tk | None = None
        self.canvas: tk.Canvas | None = None

    def on_key_press(self, e: tk.Event):
        """event callback for key presses"""
        if e.char == "-":
            self.video_container.fps_multiplier = max(self.video_container.fps_multiplier / 2, 0.25)
            logger.info(f"Speed: {self.video_container.fps_multiplier}x")
        if e.char == "=":
            self.video_container.fps_multiplier = min(self.video_container.fps_multiplier * 2, 4.0)
            logger.info(f"Speed: {self.video_container.fps_multiplier}x")
        if e.keycode == 9: # esc
            self.root.destroy()
        if e.keycode == 65: # space
            self.video_container.is_paused = not self.video_container.is_paused
        if e.keycode == 113: # left arrow -> jump half a second back
            logger.debug((self.video_container.current_frame,
                          round(self.video_container.current_frame - self.video_container.fps // 3)))
            self.video_container.current_frame = self.video_container.current_frame - self.video_container.fps // 3
        if e.keycode == 114: # right arrow -> jump half a second forward
            logger.debug((self.video_container.current_frame,
                          round(self.video_container.current_frame + self.video_container.fps // 3)))
            self.video_container.current_frame = self.video_container.current_frame + self.video_container.fps // 3
        if e.char == "i":
            self.show_stats = not self.show_stats

    def run(self):
        """main loop of the window thread"""
        self.root = tk.Tk()
        self.root.title(self.video_container.video.path)
        self.canvas = tk.Canvas(self.root, width=self.video_container.frame_shape[1],
                                height=self.video_container.frame_shape[0])
        self.canvas.pack()
        self.canvas.bind("<KeyPress>", self.on_key_press)
        self.canvas.focus_set()

        while True:
            self.root.update()
            now = datetime.now()
            try:
                img_arr = self.queue.get(block=True, timeout=TIMEOUT_READ_S * 10**(self.frame_counter == 0)) # * 10s
            except Empty:
                logger.warning("Empty queue. Exitting")
                self.root.destroy()
                break
            if self.show_stats:
                avg_fps = np.nanmean(self.fps_counter) if self.frame_counter > 0 else 0
                text = ""
                if self.video_container.current_frame != 0:
                    text = f"Frame: {self.video_container.current_frame}. "
                text += f"Avg fps: {avg_fps:.2f}"
                img_arr = image_add_text(img_arr, text=text, position=(0, 18),
                                         font_size_px=self.video_container.font_size,
                                         font_color=self.video_container.font_color)

            photo = ImageTk.PhotoImage(image=Image.fromarray(img_arr))
            self.canvas.create_image(0, 0, anchor=tk.NW, image=photo)
            self.frame_counter += 1
            if self.show_stats and not self.video_container.is_paused:
                took = (datetime.now() - now).total_seconds()
                self.fps_counter[self.frame_counter % len(self.fps_counter)] = 1 / took

def read_loop(video_container: VideoContainer, q: Queue, stats: list[str], stats_path: str | None):
    """simply reads the current frame as given by video.current_frame and writes to the queue."""
    video = video_container.video
    while True:
        try:
            q.put(video[video_container.current_frame], timeout=TIMEOUT_WRITE_S)
        except Exception as e:
            logger.info(f"Program closed: {e}")
            sys.exit(0)
        if stats_path is not None:
            stats.append(datetime.now().replace(tzinfo=None).isoformat())
            if len(stats) % 100 == 0 and len(stats) > 0:
                with open(stats_path, "w" if len(stats) == 100 else "a") as fp:
                    fp.write("frame,timestmap\n"+"\n".join(f"{i},{row}" for i, row in enumerate(stats)))
        if video_container.is_paused:
            next_frame = video_container.current_frame
        else:
            next_frame = (video_container.current_frame + 1) % len(video)
        video_container.current_frame = next_frame
        time.sleep(1 / (video.fps * video_container.fps_multiplier))

def get_args() -> Namespace:
    """cli args"""
    parser = ArgumentParser()
    parser.add_argument("video_path")
    parser.add_argument("--input_size", nargs=2, type=int, help="Must be set if video_path is a socket or stdin")
    parser.add_argument("--fps", type=float, help="Must be set if video_path is a socket or stdin. If webcam, use 1.")
    parser.add_argument("--stats_path", help="If set, it will output a per-frame stats file with timestamps for fps")
    parser.add_argument("--hud_font_size", type=int, help="Size in pixels. If not set, will use a default")
    parser.add_argument("--hud_font_color", help="R G B tuple. If not set, defaults to white.", nargs=3, type=int)
    args = parser.parse_args()
    if args.video_path == "-":
        assert args.input_size is not None, "--input_size must be set when reading frames from socket/stdin"
        assert args.fps is not None, "--fps must be set when reading frames from socket/stdin"
    return args

def main(args: Namespace):
    """main fn"""
    q = Queue(maxsize=1)
    reader_kwargs = {}
    if args.video_path == "-":
        reader_kwargs = {"resolution": args.input_size, "fps": args.fps, "async_worker": True}
    video_container = VideoContainer(VREVideo(args.video_path, **reader_kwargs), font_size=args.hud_font_size,
                                     font_color=args.hud_font_color)
    VideoPlayer(q, video_container).start()
    stats = None if args.stats_path is None else []
    read_loop(video_container, q, stats, args.stats_path)

if __name__ == "__main__":
    main(get_args())



"""utils"""
from pathlib import Path
from typing import Any, Callable, T, Iterable
from functools import partial
from urllib.request import urlretrieve
import os
import numpy as np
from tqdm import tqdm
from loggez import make_logger
from PIL import Image, ImageDraw, ImageFont

logger = make_logger("VRE_VIDEO")

def image_write(file: np.ndarray, path: str):
    """PIL image writer"""
    assert file.min() >= 0 and file.max() <= 255
    img = Image.fromarray(file.astype(np.uint8), "RGB")
    img.save(path)

def image_read(path: str) -> np.ndarray:
    """PIL image reader"""
    # TODO: for grayscale, this returns a RGB image too
    img_pil = Image.open(path)
    img_np = np.array(img_pil, dtype=np.uint8)
    # grayscale -> 3 gray channels repeated.
    if img_pil.mode == "L":
        return np.repeat(img_np[..., None], 3, axis=-1)
    # RGB or RGBA
    return img_np[..., 0:3]

def image_add_text(image_np: np.ndarray, text: str, position: tuple[int, int], font_size_px: int,
                   font_color: tuple[int, int, int] = (0, 0, 0)) -> np.ndarray:
    """Adds text to a NumPy image array at the specified position with the specified font size."""
    image = Image.fromarray(image_np)
    draw = ImageDraw.Draw(image)
    try:
        font = ImageFont.truetype("arial.ttf", font_size_px)
    except IOError:
        font = ImageFont.load_default(font_size_px)
    draw.text(position, text, font=font, fill=font_color)
    return np.array(image)

def fetch(url: str, dst: str):
    """fetches data and stores locally with pbar"""
    assert not Path(dst).exists(), f"'{dst}' exists (or is a dir). Remove first or provide destination file path + name"
    class DownloadProgressBar(tqdm):
        """requests + tqdm"""
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs, disable=os.getenv("VRE_VIDEO_PBAR", "1") == "0")

        def update_to(self, b=1, bsize=1, tsize=None):
            """Callback from tqdm"""
            if tsize is not None:
                self.total = tsize
            self.update(b * bsize - self.n)

    with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=Path(dst).name) as t:
        try:
            urlretrieve(url, filename=dst, reporthook=t.update_to)
        except Exception as e:
            logger.info(f"Failed to download '{url}' to '{dst}'")
            raise e

def parsed_str_type(item: Any) -> str:
    """Given an object with a type of the format: <class 'A.B.C.D'>, parse it and return 'A.B.C.D'"""
    return str(type(item)).rsplit(".", maxsplit=1)[-1][0:-2]

def natsorted(seq: Iterable[T], key: Callable[[T], "SupportsGTAndLT"] | None = None, reverse: bool=False) -> list[T]:
    """wrapper on top of natsorted so we can properly remove it"""
    def _try_convert_to_num(x: str) -> str | int | float:
        try:
            return int(x)
        except ValueError:
            try:
                return float(x)
            except ValueError:
                return x

    def natsorted_key(item: T, key: Callable) -> "SupportsGTAndLT":
        item = key(item)
        if isinstance(item, str):
            ix_dot = item.rfind(".")
            item = item[0:ix_dot] if ix_dot != -1 else item
            item = _try_convert_to_num(item)
        return item

    key = key or (lambda item: item)
    return sorted(seq, key=partial(natsorted_key, key=key), reverse=reverse)



"""pil_frame_writer - Module that implements frame writing using Pillow"""
from pathlib import Path
import os
from tqdm import trange

from .frame_writer import FrameWriter
from ..utils import image_write

class PILFrameWriter(FrameWriter):
    """"PILFrameWriter implementation. Writes videos to disk using PIL."""
    def __init__(self, fmt: str = "png"):
        super().__init__()
        assert fmt in {"png", "jpg"}, fmt
        self.suffix = fmt

    def write(self, video: "VREVideo", out_path: str | Path, start_frame: int = 0, end_frame: int | None = None):
        out_path = Path(out_path)
        out_path.mkdir(exist_ok=True, parents=True)
        assert len(files := list(out_path.iterdir())) == 0, f"Data exists in out_path. N files: {len(files)}"
        start_frame = start_frame or 0
        end_frame = end_frame or len(video)
        assert start_frame >= 0 and end_frame <= len(video), (start_frame, end_frame)
        for i in trange(start_frame, end_frame, disable=os.getenv("VRE_VIDEO_PBAR", "1") == "0",
                        desc=f"[PILFrameWriter] {out_path}"):
            image_write(video[i], f"{out_path}/{i}.{self.suffix}")
